package de.tum.in.net.WSNDataFramework.Modules.IPFIX;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.tum.in.net.WSNDataFramework.Node;
import de.tum.in.net.WSNDataFramework.Node.Datum.DatumID;
import de.tum.in.net.WSNDataFramework.WSNAggregationNode;
import de.tum.in.net.WSNDataFramework.WSNModule;
import de.tum.in.net.WSNDataFramework.Modules.FileLogger.WSNFileLoggerModule;
import de.tum.in.net.WSNDataFramework.Modules.HTTPServer.WSNHTTPServerModule;




/**
 * Module handling data coming from an (tiny)IPFIX based WSN. Listening to UDP socket.
 * Handles IPFIX messages as well as tinyIPFIX ones.
 * 
 * @author André Freitag
 *
 */
public class WSNTinyIPFIXModule extends WSNModule {

	/* constructors  */
	/**
	 * listen to :port, use pathToMetdata to parse packets
	 * 
	 * @param pathToMetadata path to metadata file used to translate incoming packets
	 * @param port to listen to
	 * @param address to listen to
	 * @throws Exception
	 */
	public WSNTinyIPFIXModule(String pathToMetadata, int port) throws Exception {
		// create socket
		_serverSocket = new DatagramSocket(port);

		// create parses instance
		_parser = new TinyIPFIXParser();
		_enricher = new IPFIXEnricher(pathToMetadata);

		_setIdling("no data received yet");
	}
	/**
	 * listen to :4739, use pathToMetdata to parse packets
	 * 
	 * @param pathToMetadata path to metadata file used to translate incoming packets
	 * @param port to listen to
	 * @throws Exception
	 */
	public WSNTinyIPFIXModule(String pathToConf) throws Exception {
		this(pathToConf, 4739);
	}


	/* public methods */
	//	/**
	//	 * appends a message to the log
	//	 *
	//	 * @param msg
	//	 * @return this for fluent interface
	//	 */
	//	public WSNTinyIPFIXModule log(String msg) {
	//		_parser.log(msg);
	//		return this;
	//	}
	//	public WSNTinyIPFIXModule clearLog() {
	//		_parser.clearLog();
	//		return this;
	//	}
	/**
	 * gets logged messages
	 */
	public String getLog() {
		return _parser.getLog();
	}

	/* overridden methods */
	/**
	 * name
	 */
	@Override
	public String getName() {
		return "TinyIPFIX Listener";
	}

	/**
	 * init handler
	 */
	@Override
	protected void _init() {
		_moduleDependent(WSNHTTPServerModule.class, "_moduleDep", "_moduleRem");
	}

	/**
	 * shutdown handler
	 */
	@Override
	protected void _shutdown() {
		_serverSocket.close();
	}


	/**
	 * worker thread
	 */
	@Override
	protected void _run() {

		try {
			while(!Thread.interrupted()){

				/* receive datagram */
				DataPacket p;
				try {
					p = _receiveDataPacket();
				} catch (Exception e) {
					return;
				}


				/* parse datagram */
				if (p!=null && p.data.length != 0) {

					try {
						// parse packet
						Map<String,List<IPFIXField>> fields = _parser.parse(p.data, p.address);

						/* update WSN data if fields were received.. */
						if (fields != null) {
							Map<String,Node> nodes = new HashMap<String,Node>();

							// generate nodes, fill nodes list
							for (String nodeID: fields.keySet()) {

								// get field prefixes of aggregated nodes
								Map<String,Integer> aggregatedNodesPrefixes = new HashMap<String,Integer>();
								for (IPFIXField field: fields.get(nodeID)) {
									IPFIXEnrichedField enrichedField = _enricher.enrich(field);
									if (enrichedField.type!=null && enrichedField.type.equals("NodeID") && enrichedField.value!=null && !enrichedField.value.toString().equals(nodeID)) {
										aggregatedNodesPrefixes.put(enrichedField.value.toString(), enrichedField.templateField.fieldID & 0xFF00);
									}
								}

								// add current node to list
								Node node = this.app().wsn().node(nodeID); // use a copy of the WSN's node if it already exists..
								if (node == null) {
									if (aggregatedNodesPrefixes.size() > 0) { // create new aggregated node if aggregatedNodesPrefixes is not empty
										node = new WSNAggregationNode(nodeID);
									} else { // otherwise create new simple node
										node = new Node(nodeID);
									}
									this.app().wsn().addNode(node);
								} else if(aggregatedNodesPrefixes.size() > 0 && !(node instanceof WSNAggregationNode)) { // if node was already present in current WSN as simple node but it is actually an aggregation node, convert
									this.app().wsn().removeNode(node.getNodeID());
									node = new WSNAggregationNode(node);
									this.app().wsn().addNode(node);
								} else if (aggregatedNodesPrefixes.size() == 0 && !(node instanceof Node)) {
									this.app().wsn().removeNode(node.getNodeID());
									node = new Node(node);
									this.app().wsn().addNode(node);
								}
								nodes.put(nodeID, node);
								// add aggregated nodes to list
								if (aggregatedNodesPrefixes.size() > 0) {
									WSNAggregationNode n = (WSNAggregationNode) node;
									List<Node> aggregatedNodes = new ArrayList<Node>();

									for (String aggregatedNodeId: aggregatedNodesPrefixes.keySet()) {
										if (this.app().wsn().node(aggregatedNodeId) == null) { // create new one if node is not present in current WSN
											Node newNode = new Node(aggregatedNodeId);
											this.app().wsn().addNode(newNode);
											nodes.put(aggregatedNodeId, newNode);
										} else { // otherwise use the WSN's one
											nodes.put(aggregatedNodeId, this.app().wsn().node(aggregatedNodeId));
										}

										aggregatedNodes.add(nodes.get(aggregatedNodeId));
									}

									n.updateAggregatedNodes(aggregatedNodes); // update the aggregator's aggregatedNodes list
								}


								// enrich created nodes with the received fields
								for (IPFIXField field: fields.get(nodeID)) {

									IPFIXEnrichedField enrichedField = _enricher.enrich(field);
									Node.Datum wsnNodeField = new Node.Datum(
											enrichedField.templateField.getQualifier(),
											enrichedField.name,
											enrichedField.type,
											enrichedField.value,
											enrichedField.unit
											);

									boolean added=false;
									if (aggregatedNodesPrefixes.size() > 0) {
										for (String aggregatedNodeId: aggregatedNodesPrefixes.keySet()) {
											Integer aggregatedPrefix = aggregatedNodesPrefixes.get(aggregatedNodeId);

											if ((enrichedField.templateField.fieldID & 0xFF00) == aggregatedPrefix) { // field belongs to current aggregatedNodeId
												wsnNodeField = new Node.Datum(
														new DatumID(enrichedField.templateField.enterpriseNumber+"|"+(enrichedField.templateField.fieldID & 0x00FF)),
														enrichedField.name,
														enrichedField.type,
														enrichedField.value,
														enrichedField.unit
														);
												System.out.println("update aggr!");
												nodes.get(aggregatedNodeId).data().update(wsnNodeField);
												added=true;
											}
										}
									}

									if (!added) {
										nodes.get(nodeID).data().update(wsnNodeField);
									}
								}
							}


							// adapt status
							if (this.getStatus().getStatus() != WSNModuleStatus.STATUS.RUNNING) {
								_setRunning("up and running");
							}
						}
					} catch (TinyIPFIXParser.ParseException e) {
						System.err.println("COULDN'T PARSE IPFIX PACKET:");
						e.printStackTrace();
					} catch (Exception e) {
						e.printStackTrace();
					}


				} else { // 0 bytes received.. socket broken.
					break;
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			// clean up..
			_serverSocket.close();
		}
	}

	protected void _moduleDep(WSNHTTPServerModule m) {
		try {
			m.registerController("ipfix", HTTPController.class, this)
			.getServerModule().registerLink(new String[]{"Visualisation","TinyIPFIX"}, "/ipfix");
		} catch (InstantiationException e) {
		}
	}

	/* protected helper	 */
	/**
	 * receive another data packet. blocks thread until a new packet is available.
	 *
	 * @return data packet
	 * @throws IOException
	 */
	protected DataPacket _receiveDataPacket() throws IOException {
		if (_serverSocket.isClosed()) return null;

		byte[] buf = new byte[2048];
		DatagramPacket p = new DatagramPacket(buf, buf.length);
		_serverSocket.receive(p);

		byte[] data = new byte[p.getLength()];
		System.arraycopy(p.getData(), 0, data, 0, data.length);

		// log IPFIX packets
		if (this.app().module(WSNFileLoggerModule.class) != null) {

			String dataHex = toHexString(data);
			dataHex +="\n";

			File log = this.app().module(WSNFileLoggerModule.class).getLogFile("listener-capture-hex.txt");

			FileOutputStream fos = new FileOutputStream(log,true);
			fos.write(dataHex.getBytes());
			fos.close();
		}

		return new DataPacket(data, p.getAddress());
	}
	public static String toHexString(byte[] bytes) {
		char[] hexArray = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
		char[] hexChars = new char[bytes.length * 2];
		int v;
		for ( int j = 0; j < bytes.length; j++ ) {
			v = bytes[j] & 0xFF;
			hexChars[j*2] = hexArray[v/16];
			hexChars[j*2 + 1] = hexArray[v%16];
		}
		return new String(hexChars);
	}
	/**
	 * class representing the result of _receiveDataPacket()
	 * @author André Freitag
	 */
	protected class DataPacket {
		public DataPacket(byte[] data, InetAddress address) {
			this.data = data;
			this.address = address;
		}
		InetAddress address;
		byte[] data;
	}


	//	/* protected members */
	protected DatagramSocket _serverSocket;
	protected TinyIPFIXParser _parser=null;
	protected IPFIXEnricher _enricher=null;
}
